import numpy as np
import scipy as sp
from sklearn.decomposition import PCA
from sklearn.mixture import GaussianMixture

from gum import GaussianUniformMixture


def compute_pca(training_features,
                n_components,
                *other_features,
                whiten=True,
                return_pca=False):
    pca = PCA(whiten=whiten, n_components=n_components)
    training_components = pca.fit_transform(training_features)
    other_components = [pca.transform(feature) for feature in other_features]
    if return_pca:
        return (training_components, *other_components, pca)
    else:
        return (training_components, *other_components)


def predict_direct(features_weighted,
                   features_full,
                   training_mask,
                   pca_n_components=25):
    """
    Given an image and a suspicious region,
    decompose the image in principal components and computes
    the Mahalanobis distance of the features of each block
    to the distribution of non-suspicious blocks
    :input features: array-like, shape (Y, X, c)
    :input sat_mask: bool array-like, shape (Y, X). Masks saturated blocks.
    :input suspect_mask: bool array-like, shape (Y, X). Suspicious region.
    :param pca_n_components: int < c, default: 25. None: keep all.
    :return mahalanobis: np.ndarray, shape (Y, X)
    """
    Y, X, c = features_weighted.shape
    list_features_valid = features_weighted[training_mask].reshape(-1, c)
    list_features_weighted = features_weighted.reshape(-1, c)
    list_features_full = features_full.reshape(-1, c)
    pca_n_components = min(pca_n_components, min(list_features_valid.shape))
    pca_valid, pca_weighted, pca_full = compute_pca(list_features_valid,
                                                    pca_n_components,
                                                    list_features_weighted,
                                                    list_features_full)
    pca_mean = pca_valid.mean(axis=0)  # shape (1, pca_n_components)
    mahalanobis_weighted = np.sqrt(
        np.sum(np.square(pca_weighted - pca_mean[None]),
               axis=1)).reshape(Y, X)
    mahalanobis_full = np.sqrt(
        np.sum(np.square(pca_full - pca_mean[None]), axis=1)).reshape(Y, X)
    mahalanobis_valid = mahalanobis_weighted * training_mask
    return mahalanobis_valid, mahalanobis_weighted, mahalanobis_full


def predict_gaussian_mixture(
    features_weighted,
    features_full,
    training_mask,
    pca_n_components=25,
    **kwargs,
):
    """
    Given features from image blocks, decompose the features
    in principal components and fit a Gaussian Mixtures model
    with 2 components on the features to find forged regions.
    :input features: array-like, shape (Y, X, c)
    :input suspect_mask: bool array-like, shape (Y, X). Suspicious region.
    :param pca_n_components: int < c, default: 25. None: keep all.
    :**kwargs: passed to sklearn.mixture.GaussianMixture 
    :return out: np.ndarray, shape (Y, X)
    """
    Y, X, c = features_weighted.shape
    list_features_valid = features_weighted[training_mask].reshape(-1, c)
    list_features_weighted = features_weighted.reshape(-1, c)
    list_features_full = features_full.reshape(-1, c)
    pca_n_components = min(pca_n_components, min(list_features_valid.shape))
    pca_valid, pca_weighted, pca_full = compute_pca(list_features_valid,
                                                    pca_n_components,
                                                    list_features_weighted,
                                                    list_features_full)
    gm = GaussianMixture(n_components=2, **kwargs)
    gm.fit(pca_valid)
    Σ = gm.covariances_
    μ = gm.means_

    centered_weighted_0 = pca_weighted - μ[0]
    mahalanobis_weighted_0 = sp.linalg.norm(sp.linalg.solve(
        sp.linalg.cholesky(Σ[0]), centered_weighted_0.T),
                                            axis=0,
                                            ord=2)
    centered_weighted_1 = pca_weighted - μ[1]
    mahalanobis_weighted_1 = sp.linalg.norm(sp.linalg.solve(
        sp.linalg.cholesky(Σ[1]), centered_weighted_1.T),
                                            axis=0,
                                            ord=2)
    # Choose the authentic and forged distributions
    mahalanobis_weighted = mahalanobis_weighted_0 if np.median(
        mahalanobis_weighted_0) < np.median(
            mahalanobis_weighted_1) else mahalanobis_weighted_1
    mahalanobis_weighted = mahalanobis_weighted.reshape(Y, X)

    mahalanobis_valid = mahalanobis_weighted_0 * training_mask if np.median(
        mahalanobis_weighted_0[training_mask]) < np.median(
            mahalanobis_weighted_1[training_mask]
        ) else mahalanobis_weighted_1 * training_mask

    centered_full_0 = pca_full - μ[0]
    mahalanobis_full_0 = sp.linalg.norm(sp.linalg.solve(
        sp.linalg.cholesky(Σ[0]), centered_full_0.T),
                                        axis=0,
                                        ord=2)
    centered_full_1 = pca_full - μ[1]
    mahalanobis_full_1 = sp.linalg.norm(sp.linalg.solve(
        sp.linalg.cholesky(Σ[1]), centered_full_1.T),
                                        axis=0,
                                        ord=2)
    # Choose the authentic and forged distributions
    mahalanobis_full = mahalanobis_full_0 if np.median(
        mahalanobis_full_0) < np.median(
            mahalanobis_full_1) else mahalanobis_full_1
    mahalanobis_full = mahalanobis_full.reshape(Y, X)

    # γ = gm.predict_proba(pca_all)
    # γ = γ[:,
    # 0] if np.median(mahalanobis_0) < np.median(mahalanobis_1) else γ[:,
    # 1]
    # γ = γ.reshape(Y, X)
    # return mahalanobis, γ
    return mahalanobis_valid, mahalanobis_weighted, mahalanobis_full


def predict_gaussian_uniform(features_weighted,
                             features_full,
                             training_mask,
                             pca_n_components=25,
                             **kwargs):
    Y, X, c = features_weighted.shape
    list_features_valid = features_weighted[training_mask].reshape(-1, c)
    list_features_weighted = features_weighted.reshape(-1, c)
    list_features_full = features_full.reshape(-1, c)
    pca_n_components = min(pca_n_components, min(list_features_valid.shape))
    pca_valid, pca_weighted, pca_full = compute_pca(list_features_valid,
                                                    pca_n_components,
                                                    list_features_weighted,
                                                    list_features_full)
    gum = GaussianUniformMixture(**kwargs)
    gum.fit(pca_valid)
    mahalanobis_weighted = gum.predict(pca_weighted)[0].reshape(Y, X)
    mahalanobis_valid = mahalanobis_weighted * training_mask
    mahalanobis_full = gum.predict(pca_full)[0].reshape(Y, X)
    # γ = γ.reshape(Y, X)
    # return mahalanobis, γ
    return mahalanobis_valid, mahalanobis_weighted, mahalanobis_full
