from functools import reduce

import numpy as np
import scipy as sp
from skimage import morphology


def normalize(img):
    """
    Converts an image to float in the [0, 1] range
    :input img: array-like
    :return img: float np.ndarray
    """
    img = np.asarray(img)
    if img.ndim == 3:
        img = img[:, :, :3]  # discard eventual alpha channel
    if img.dtype.kind == 'f':  # Checks for float or any np.float* type
        return img
    elif img.dtype == np.uint8:
        return img / (2**8 - 1)
    elif img.dtype == np.uint16:
        return img / (2**16 - 1)
    elif img.dtype == np.uint32:
        return img / (2**32 - 1)
    elif img.dtype == np.uint64:
        return img / (2**64 - 1)
    else:
        raise ValueError(img.dtype)


def mask_saturated_regions(img,
                           threshold_low=6 / 255,
                           threshold_high=252 / 255):
    """
    Computes a mask of the saturated regions in an image.
    For each channel, pixels below `threshold_low` are marked.
    Morphological opening is performed on the mask,
    with a neighbourhood of a radius-3 disk.
    Another mask is computed for pixels above `threshold_high`,
    this time with a radius-2 disk as neighbourhood for the opening.
    The two masks are then merged across all directions to yield only one mask,
    which is morphological dilation with a 9x9 square as neighbourhood.
    We invert the mask before returning,
    so that the output shows correct regions and masks saturated ones.
    The original paper uses 6, 252 as thresholds for uint8 images
    :input img: array-like, shape (Y, X(, C))
    :param threshold_low: scalar
    :param threshold_high: scalar
    :return mask: np.ndarray, shape (Y, X)
    """
    if img.ndim == 2:
        img = img[:, :, None]
    img = img.transpose(2, 0, 1)
    masks_low = [
        morphology.binary_opening(channel, morphology.disk(3))
        for channel in img < threshold_low
    ]
    masks_high = [
        morphology.binary_opening(channel, morphology.disk(2))
        for channel in img > threshold_high
    ]
    mask = reduce(np.logical_or, masks_low + masks_high)
    mask = morphology.binary_dilation(mask, np.ones((9, 9)))
    mask = ~mask
    return mask


def to_grayscale(img):
    """
    Converts an image to grayscale.
    If the image is already grayscale, returns it untouched.
    :input img: array-like, shape (Y, X(, C))
    :return img: np.ndarray, shape (Y, X)
    """
    if img.ndim == 2:
        return img
    else:
        img = (0.299 * img[:, :, 0] + 0.587 * img[:, :, 1] +
               0.114 * img[:, :, 2])
        return img


def high_pass(img):
    """
    Correlates the image with high-pass filter [0, 1, -3, 3, -1].
    Correlation is done in both directions.
    :input img: array-like, shape (Y, X)
    :return residual_h: np.ndarray, shape (Y-4, X-4). Horizontal residual
    :return residual_v: np.ndarray, shape (Y-4, X-4). Vertical residual
    """
    kernel = np.array([[0., 1., -3., 3., -1.]])
    residual_h = sp.signal.correlate2d(img, kernel, mode='same')[2:-2, 2:-2]
    residual_v = sp.signal.correlate2d(img, kernel.T, mode='same')[2:-2, 2:-2]
    return residual_h, residual_v


def quantize_and_clip(residual, z=6 / 255):
    """
    The original paper divides the residual by a quantization factor z,
    then rounds it to the nearest integer, adds 1 to it,
    and clips it to [0, 2].
    This is equivalent to defining the output such as
    * y[x] = 0 if residual[x] < -z/2
    * y[x] = 1 if -z/2 ≤ residual[x] ≤ z/2
    * y[x] = 2 if z/2 < residual[x]
    z thus represents the length of the center bin.
    The original paper uses the value 6/255
    :input residual: array-like
    :param z: scalar, default: 6/255.
    :return q: int np.ndarray, same shape as residual. Quantized residual
    """
    q = np.ones(residual.shape, int)
    q[residual < -z / 2] = 0
    q[residual > z / 2] = 2
    return q


def _reduce_cooccurence_matrix(cooc, symmetry=True, complementary=True):
    # TODO move reduction for all matrices simultaneously
    """
    Reduces a co-occurrence matrix as computed below.
    The input list contains the [0, 2] int values of each pixels.
    The 81 possible combinations are reduced to 25 by mapping
    (a, b, c, d), (d, c, b, a), (2-a, 2-b, 2-c, 2-d) and (2-d, 2-c, 2-b, 2-a)
    to the same entry.
    :input cooc: [0, 2] int np.ndarray, shape (4, Y, X)
    :return reduced: [0, 24] int np.ndarray, shape (Y, X)
    """
    eastward = 27 * cooc[0] + 9 * cooc[1] + 3 * cooc[2] + cooc[3]
    indices = eastward
    if symmetry:
        westward = cooc[0] + 3 * cooc[1] + 9 * cooc[2] + 27 * cooc[3]
        indices = np.minimum(eastward, westward)
        if complementary:
            indices = np.minimum(indices, 80 - westward)
    if complementary:
        indices = np.minimum(indices, 80 - eastward)
    # indices shape: (Y, X)
    _, reduced = np.unique(indices, return_inverse=True)
    # np.unique flattens the array, we need to reshape it
    reduced = reduced.reshape(indices.shape)
    return reduced


def cooccurence_matrices(residual, **kwargs):
    """
    Computes the co-occurence matrices of a quantized residual in [0, 2].
    Each value in the matrices identify the tuple formed
    by the pixel and the three pixels rightward/downward,
    up to a symmetry and/or a complement to 2.
    :input residual: int array-like, shape (Y, X), with values in [0, 2]
    :return cooc_h: int [0, 24] np.ndarray, shape (Y-4, X-4). Horizontal co-occurence matrix.
    :return cooc_v: int [0, 24] np.ndarray, shape (Y-4, X-4). Vertical co-occurence matrix.
    """
    cooc_h = np.stack([
        residual[2:-2, 0:-4], residual[2:-2, 1:-3], residual[2:-2, 2:-2],
        residual[2:-2, 3:-1]
    ])
    cooc_v = np.stack([
        residual[0:-4, 2:-2], residual[1:-3, 2:-2], residual[2:-2, 2:-2],
        residual[3:-1, 2:-2]
    ])
    cooc_h = _reduce_cooccurence_matrix(cooc_h, **kwargs)
    cooc_v = _reduce_cooccurence_matrix(cooc_v, **kwargs)
    return cooc_h, cooc_v


def compute_histograms(cooc_h_res_h,
                       cooc_v_res_h,
                       cooc_h_res_v,
                       cooc_v_res_v,
                       mask,
                       stride,
                       block_size,
                       pooling='half',
                       valid_pixels_min_ratio=0.85,
                       normalization='correct'):
    Y, X = cooc_h_res_h.shape
    img_size = (Y + 8) * (X + 8)
    mask = mask[4:-4, 4:-4]
    if pooling == 'half':
        n_histograms = 2
    elif pooling == 'full':
        n_histograms = 1
    elif pooling == 'none':
        n_histograms = 4
    else:
        raise ValueError(f'{pooling} must be one of "full", "half", "none".')

    assert normalization in ['correct', 'original']

    n_bins = 1 + max(
        max(max(cooc_h_res_h.max(), cooc_h_res_v.max()), cooc_v_res_h.max()),
        cooc_v_res_v.max())

    n_features = n_histograms * n_bins
    cuts_y = np.arange(0, Y - stride + 1, stride)
    cuts_x = np.arange(0, X - stride + 1, stride)
    bins = np.arange(0, n_bins + 1)

    n_Y = cuts_y.size  # number of windows in each direction
    n_X = cuts_x.size

    histograms_weighted = np.zeros((n_Y, n_X, n_features))
    histograms_full = np.zeros((n_Y, n_X, n_features))
    window_weights = np.zeros((n_Y, n_X))

    for i_y, y in enumerate(cuts_y):
        for i_x, x in enumerate(cuts_x):
            this_weights = mask[y:y + stride, x:x + stride]
            this_weight_total = this_weights.sum()
            window_weights[i_y, i_x] = this_weight_total / (stride**2)

            if normalization == 'original' and this_weight_total > 0:
                this_weights /= this_weight_total

            hist_h_h_w = np.histogram(cooc_h_res_h[y:y + stride, x:x + stride],
                                      bins,
                                      weights=this_weights)[0].astype(float)
            hist_h_h_f = np.histogram(cooc_h_res_h[y:y + stride, x:x + stride],
                                      bins)[0].astype(float)
            hist_v_h_w = np.histogram(cooc_v_res_h[y:y + stride, x:x + stride],
                                      bins,
                                      weights=this_weights)[0].astype(float)
            hist_v_h_f = np.histogram(cooc_v_res_h[y:y + stride, x:x + stride],
                                      bins)[0].astype(float)
            hist_h_v_w = np.histogram(cooc_h_res_v[y:y + stride, x:x + stride],
                                      bins,
                                      weights=this_weights)[0].astype(float)
            hist_h_v_f = np.histogram(cooc_h_res_v[y:y + stride, x:x + stride],
                                      bins)[0].astype(float)
            hist_v_v_w = np.histogram(cooc_v_res_v[y:y + stride, x:x + stride],
                                      bins,
                                      weights=this_weights)[0].astype(float)
            hist_v_v_f = np.histogram(cooc_v_res_v[y:y + stride, x:x + stride],
                                      bins)[0].astype(float)

            if normalization == 'original' and this_weight_total > 0:
                hist_h_h_w /= hist_h_h_w.sum()
                hist_v_h_w /= hist_v_h_w.sum()
                hist_h_v_w /= hist_h_v_w.sum()
                hist_v_v_w /= hist_v_v_w.sum()
                hist_h_h_f /= hist_h_h_f.sum()
                hist_v_h_f /= hist_v_h_f.sum()
                hist_h_v_f /= hist_h_v_f.sum()
                hist_v_v_f /= hist_v_v_f.sum()

            if pooling == 'half':
                histograms_weighted[i_y, i_x] = np.concatenate(
                    [hist_h_v_w + hist_v_h_w, hist_h_h_w + hist_v_v_w]) / 2
                histograms_full[i_y, i_x] = np.concatenate(
                    [hist_h_v_f + hist_v_h_f, hist_h_h_f + hist_v_v_f]) / 2
            elif pooling == 'full':
                histograms_weighted[i_y, i_x] = (hist_h_v_w + hist_v_h_w +
                                                 hist_h_h_w + hist_v_v_w) / 4
                histograms_full[i_y, i_x] = (hist_h_v_f + hist_v_h_f +
                                             hist_h_h_f + hist_v_v_f) / 4
            else:
                histograms_weighted[i_y, i_x] = np.concatenate(
                    [hist_h_v_w, hist_v_h_w, hist_h_h_w, hist_v_v_w])
                histograms_full[i_y, i_x] = np.concatenate(
                    [hist_h_v_f, hist_v_h_f, hist_h_h_f, hist_v_v_f])

    sum_window_size = block_size // stride
    sum_window_filter = np.ones((sum_window_size, sum_window_size), float)

    features_weighted = np.zeros(
        (n_Y - sum_window_size + 1, n_X - sum_window_size + 1, n_features),
        float)
    features_full = np.zeros(
        (n_Y - sum_window_size + 1, n_X - sum_window_size + 1, n_features),
        float)
    summed_weights = sp.signal.correlate2d(window_weights,
                                           sum_window_filter,
                                           mode='valid')
    for i in range(n_features):
        features_weighted[:, :,
                          i] = sp.signal.correlate2d(histograms_weighted[:, :,
                                                                         i],
                                                     sum_window_filter,
                                                     mode='valid')
        features_full[:, :, i] = sp.signal.correlate2d(histograms_full[:, :,
                                                                       i],
                                                       sum_window_filter,
                                                       mode='valid')
    if normalization == 'correct':
        # Do the full normalization at the end
        features_weighted /= n_histograms * stride**2 * np.maximum(
            summed_weights[:, :, None], 1e-20)
        features_full /= np.maximum(features_full.sum(axis=-1, keepdims=True),
                                    1e-20)
    else:
        features_weighted /= n_histograms * sum_window_size**2 * np.maximum(
            summed_weights[:, :, None] / (sum_window_size**2), 1e-20)
        features_full /= n_histograms * sum_window_size**2

    # keep pixels that are not in a heavily saturated region, discard the rest.
    mask = summed_weights / (sum_window_size**2) > valid_pixels_min_ratio
    mask_loose = summed_weights > 0  # only for display

    # Compensate the borders from high-pass filtering and
    # co-occurence matric computation,
    # then move to the centre of each window
    cuts_y = cuts_y + 4 + (stride - 1) / 2
    cuts_x = cuts_x + 4 + (stride - 1) / 2
    # Compensate the border from moving window average,
    # and again adjust to the centre of each block
    low = int(np.floor((sum_window_size - 1) / 2))
    high = int(np.ceil((sum_window_size - 1) / 2))
    cuts_y = (cuts_y[low:-high] + cuts_y[high:-low]) / 2
    cuts_x = (cuts_x[low:-high] + cuts_x[high:-low]) / 2
    return features_weighted, features_full, mask, mask_loose, cuts_y, cuts_x


def compute_splicebuster_features(img,
                                  quantization_factor=6 / 255,
                                  cooc_symmetry=True,
                                  cooc_complementary=True,
                                  hist_pooling='half',
                                  stride='auto',
                                  block_size='auto',
                                  valid_pixels_min_ratio=0.85,
                                  normalization='correct',
                                  use_sqrt=True):
    img = normalize(img)
    img_gray = to_grayscale(img)
    # TODO error here if image is too small
    if stride == 'auto':
        stride = 8 if img.size >= 20000 else 4
    if block_size == 'auto':
        block_size = 128 if img.size >= 20000 else 64
    assert block_size % stride == 0
    mask = mask_saturated_regions(img).astype(float)
    residual_h, residual_v = high_pass(img_gray)
    residual_h = quantize_and_clip(residual_h, z=quantization_factor)
    residual_v = quantize_and_clip(residual_v, z=quantization_factor)
    cooc_res_h = cooccurence_matrices(
        residual_h, symmetry=cooc_symmetry,
        complementary=cooc_complementary)  # result is a tuple
    cooc_res_v = cooccurence_matrices(
        residual_v, symmetry=cooc_symmetry,
        complementary=cooc_complementary)  # result is a tuple
    features_weighted, features_full, mask, mask_loose, cuts_y, cuts_x = compute_histograms(
        *cooc_res_h,
        *cooc_res_v,
        mask,
        stride=stride,
        block_size=block_size,
        pooling=hist_pooling,
        normalization=normalization)
    if use_sqrt:
        features_weighted = np.sqrt(features_weighted)
        features_full = np.sqrt(features_full)
    return features_weighted, features_full, mask, mask_loose, cuts_y, cuts_x
