import numpy as np
import scipy as sp


class GaussianUniformMixture():
    def __init__(self,
                 p_outlier_init=1e-2,
                 outlier_nlogl=42,
                 tol=1e-5,
                 max_iter=100,
                 n_init=30):
        self.p_outlier_init = p_outlier_init  # Saved for all inits
        self.outlier_nlogl = outlier_nlogl
        self.π = 1 - p_outlier_init
        self.max_iter = max_iter
        self.tol = tol
        self.n_init = n_init

    def fit(self, X):
        best_loss = np.inf
        save = None, None, None
        for i in range(self.n_init):
            loss = self._fit_once(X)
            if loss < best_loss:
                best_loss = loss
                save = self.μ, self.Σ, self.π
        self.μ, self.Σ, self.π = save

    def _fit_once(self, X):
        n_samples, n_features = X.shape
        init_index = np.random.randint(0, n_samples - 1)
        self.μ = X[init_index]
        variance = np.var(X, axis=0)
        variance += np.spacing(variance.max())
        self.Σ = np.diag(variance)
        self.π = 1 - self.p_outlier_init
        loss_old = np.inf
        for i in range(self.max_iter):
            γ, loss, _ = self._e_step(X)
            loss_diff = loss - loss_old
            if 0 <= loss_diff < self.tol * np.abs(loss):
                break
            loss_old = loss
            self._m_step(X, γ)
        return loss

    def _m_step(self, X, γ):
        n_samples, n_features = X.shape
        self.π = np.mean(γ)
        self.μ = γ.dot(X) / (n_samples * self.π)
        Xc = (X - self.μ) * np.sqrt(γ[:, None])
        self.Σ = (Xc.T @ Xc) / (n_samples * self.π) + np.spacing(
            self.Σ) * np.eye(n_features)

    def _cholesky(self, max_attempts=5):
        try:
            L = np.linalg.cholesky(self.Σ)
        except np.linalg.LinAlgError as Error:  # Σ is not positive definite
            for i in range(5):  # try regularizing it several times
                # print(
                # f"Covariance matrix is not positive definite for Cholesky decomposition. Attempt to regularize #{i}."
                # )
                w, v = sp.linalg.eigh(self.Σ)
                w = np.maximum(w, np.spacing(w.max()))
                self.Σ = v @ np.diag(w) @ v.T
                try:
                    L = np.linalg.cholesky(self.Σ)
                    break
                except np.linalg.LinAlgError:
                    continue
            else:  # if it still fails, raise an error
                raise np.linalg.LinAlgError
        return L

    def _get_nlogl(self, X):
        n_samples, n_features = X.shape
        L = self._cholesky()  # Σ = L@L.T
        D = np.diag(L)
        Xc = X - self.μ
        # Mahalanobis distance is now the L2 norm of L⁻¹ @ Xc.T
        # along the components axis
        mahalanobis = sp.linalg.norm(sp.linalg.solve(L, Xc.T), axis=0, ord=2)
        nlogl = .5 * (np.square(mahalanobis) +
                      n_features * np.log(2 * np.pi)) + np.sum(np.log(D))

        return nlogl, mahalanobis

    def _e_step(self, X):
        nlogl, mahal = self._get_nlogl(X)
        log_γ_inlier = np.log(self.π) - nlogl
        log_γ_outlier = np.log(1 - self.π) - self.outlier_nlogl
        log_γ_inlier = log_γ_inlier[:, None]
        log_γ_outlier = log_γ_outlier.repeat(log_γ_inlier.shape[0])[:, None]
        log_γ = np.append(log_γ_inlier, log_γ_outlier, axis=1)
        max_log_likelihood = np.max(log_γ, axis=1, keepdims=True)
        γ = np.exp(log_γ - max_log_likelihood)
        dem = np.sum(γ, axis=1, keepdims=True)
        γ /= dem
        loss = np.mean(np.log(dem) + max_log_likelihood)
        # equivalent to a softmax but we also compute the loss
        # γ = sp.special.softmax(log_γ, axis=1)
        return γ[:, 0], loss, mahal

    def predict(self, X):
        γ, _, mahal = self._e_step(X)
        return mahal, γ
