#!/usr/bin/env python3

import argparse
from sys import argv

import imageio
import numpy as np
import scipy as sp
import scipy.stats
import seaborn as sns
from matplotlib import colors
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

from compare_features import (predict_direct, predict_gaussian_mixture,
                              predict_gaussian_uniform)
from compute_features import compute_splicebuster_features
from postprocessing import (add_gaussian_noise, apply_median_filter,
                            gamma_correct, jpeg_compress)

FIGSIZE = (14, 14)


def get_parser():
    parser = argparse.ArgumentParser(
        description=
        'Analyze an image for forgeries with the Splicebuster algorithm.')
    parser.add_argument('input', type=str, help='Image to analyze')
    parser.add_argument(
        '-o',
        '--out',
        default='out',
        help="Path to store the outputs. Images will be saved as PATH_*.png.")

    features_group = parser.add_argument_group(
        "Arguments concerning the computation of image features")
    features_group.add_argument(
        '--quantization',
        type=int,
        default="6",
        help="Size of the center bin during quantization. Default: 6")
    cooc_group = features_group.add_argument_group(
        "Co-ocurence matrix reduction")
    sym_group = cooc_group.add_mutually_exclusive_group()
    sym_group.add_argument(
        '--symmetry',
        action='store_true',
        dest='symmetry',
        help=
        "Map symmetric entries (eg (a, b, c, d) and (d, c, b, a)) to the same value."
    )
    sym_group.add_argument(
        '--no-symmetry',
        action='store_false',
        dest='symmetry',
        help='Do not reduce the co-occurence matrix by symmetries.')
    compl_group = features_group.add_mutually_exclusive_group()
    compl_group.add_argument(
        '--complementary',
        action='store_true',
        dest='complementary',
        help=
        "Complementary-to-2 entries (eg (a, b, c, d) and (2-a, 2-b, 2-c, 2-d)) will be mapped to the same value"
    )
    compl_group.add_argument(
        '--no-complementary',
        action='store_false',
        dest='complementary',
        help='Do not reduce the co-occurence matrix by complementaries.')
    parser.set_defaults(symmetry=True, complementary=True)
    sqrt_group = features_group.add_mutually_exclusive_group()
    sqrt_group.add_argument('--sqrt',
                            action='store_true',
                            dest='sqrt',
                            help='Use the square root of the histograms.')
    sqrt_group.add_argument(
        '--no-sqrt',
        action='store_false',
        dest='sqrt',
        help='Do not perform square rooting of the histograms.')
    parser.set_defaults(sqrt=True)
    features_group.add_argument(
        '--hist-pooling',
        choices=['full', 'half', 'none'],
        default='half',
        help=
        "How to pool the histograms. `full`: All four histograms are added together, `half`: dual histograms are added, the two remaining groups are concatenated. `none`: the four histograms are concatenated."
    )
    features_group.add_argument(
        '--block-size',
        type=int,
        default=128,
        help="Block size for the moving window average")
    features_group.add_argument(
        '--stride-divider',
        type=int,
        default=16,
        help="Stride is the ratio of the block size by this number")
    features_group.add_argument(
        '--max-saturated',
        type=float,
        default=.15,
        help=
        "Maximum ratio of saturated pixels for a window to be considered valid."
    )
    histograms_normalization_group = features_group.add_mutually_exclusive_group(
    )
    histograms_normalization_group.add_argument(
        '--normalize-correct',
        action='store_const',
        const='correct',
        dest='histograms_normalization',
        help='Use the correct normalization for the histograms.')
    histograms_normalization_group.add_argument(
        '--normalize-original',
        action='store_const',
        const='original',
        dest='histograms_normalization',
        help='Normalize the histograms according to the original article.')
    parser.set_defaults(histograms_normalization='correct')
    features_group.add_argument(
        '--pca',
        type=int,
        default=5,
        help=
        'Number of components for principal components analysis. Use 0 to keep all components.'
    )

    supervision_group = parser.add_mutually_exclusive_group()
    supervision_group.add_argument(
        '--box',
        nargs=4,
        action='store',
        dest='supervision',
        help=
        'Bounding box of a suspicious region for supervised detection: X_start Y_start Width Height'
    )
    supervision_group.add_argument(
        '--mask',
        type=str,
        action='store',
        dest='supervision',
        help='Path to the mask of a suspicious region for supervised detection'
    )
    supervision_group.add_argument(
        '--gm',
        action='store_const',
        dest='supervision',
        const='gm',
        help='Unsupervised detection with a Gaussian-Gaussian Mixture')
    supervision_group.add_argument(
        '--gum',
        action='store_const',
        dest='supervision',
        const='gum',
        help='Unsupervised detection with a Gaussian-Uniform Mixture')
    supervision_group.add_argument(
        '--direct',
        action='store_const',
        dest='supervision',
        const='direct',
        help="Unsupervised detection with direct estimation of the model")
    parser.set_defaults(supervision='direct')

    em_group = parser.add_argument_group(
        "EM parameters (only used with unsupervised mixture models)")
    em_group.add_argument(
        '--n-init',
        default=5,
        type=int,
        help="Number of initializations to use (the best result will be kept)")
    em_group.add_argument(
        '--max-iter',
        default=100,
        type=int,
        help="Maximum number of iterations for each EM initialization")
    em_group.add_argument('--tolerance',
                          type=float,
                          default=1e-5,
                          help="Tolerence criterion for convergence")
    gum_group = em_group.add_argument_group(
        "GUM parameters (only used with the GUM model)")
    gum_group.add_argument('--logalpha',
                           type=float,
                           default=42,
                           help="Log outlier threshold")
    gum_group.add_argument('--p-outlier-init',
                           type=float,
                           default=1e-2,
                           help="Initial probability of outlier")
    postprocessing = parser.add_argument_group(
        "Postprocessing to test the robustness of the method")
    postprocessing.add_argument(
        "--noise",
        default=None,
        type=float,
        help='Add AWGN of the specified std (as degrees of 255) to the image')
    postprocessing.add_argument(
        "--gray",
        action='store_true',
        default=False,
        help=
        'Added AWGN will be gray. Has no effect if --noise is not provided.')
    postprocessing.add_argument("--gamma",
                                default=None,
                                type=float,
                                help="Gamma correction factor")
    postprocessing.add_argument("--jpeg",
                                default=None,
                                type=int,
                                help="JPEG compression with the given factor")
    postprocessing.add_argument("--median",
                                action='store_true',
                                default=False,
                                help="Median filtering")

    return parser


def create_colorbar(plot, ax):
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad='1%')
    cb = plt.colorbar(plot, cax=cax)
    cax.tick_params(labelsize=15)
    return cb


def _to_fullsize(cuts_y, cuts_x, data, Y, X, method='linear', fill=None):
    interpolator = sp.interpolate.RegularGridInterpolator(
        (cuts_y, cuts_x),
        data.reshape(len(cuts_y), len(cuts_x)),
        method=method,
        bounds_error=False,
        fill_value=fill)
    target_coords = np.asarray(np.meshgrid(np.arange(0, Y), np.arange(0, X)))
    result = interpolator(target_coords.reshape(2, -1).T).reshape(X, Y).T
    return result


def process_image(img, args):
    img = img[:, :, :3]
    if args.gamma is not None:
        img = gamma_correct(img, args.gamma)
    if args.median is not None:
        img = apply_median_filter(img)
    if args.noise is not None:
        img = add_gaussian_noise(img, args.noise, args.gray)
    if args.jpeg is not None:
        img = jpeg_compress(img, args.jpeg)
    return img


if __name__ == '__main__':
    parser = get_parser()
    args = parser.parse_args(argv[1:])
    img_path = args.input
    img = imageio.imread(img_path)
    if img.ndim == 2:
        img = img[:, :, None]
    img = process_image(img, args)
    Y, X = img.shape[:2]

    pca_n_components = args.pca
    supervision = args.supervision
    if supervision == 'gum':
        mode = 'mixture'
        predictor = predict_gaussian_uniform
        kwargs_mixture = {
            'p_outlier_init': args.p_outlier_init,
            'outlier_nlogl': args.logalpha
        }
    elif supervision == 'gm':
        mode = 'mixture'
        predictor = predict_gaussian_mixture
        kwargs_mixture = {}
    elif supervision == 'direct':
        mode = 'direct'
        mask = np.zeros((Y, X), bool)
    else:
        mode = 'direct'
        if type(supervision) is str:
            mask = imageio.imread(supervision).astype(bool)
            if mask.ndim == 3:
                mask = mask[:, :, 0]
            Ym, Xm = mask.shape
            if Ym != Y or Xm != X:
                raise ValueError("Image and mask must be the same size.")
        else:
            x_start, y_start, width, height = map(int, supervision)
            mask = np.zeros((Y, X), bool)
            mask[y_start:y_start + height, x_start:x_start + width] = True

    features_weighted, features_full, mask_valid, mask_loose, cuts_y, cuts_x = compute_splicebuster_features(
        img,
        quantization_factor=args.quantization / 255,
        cooc_symmetry=args.symmetry,
        cooc_complementary=args.complementary,
        hist_pooling=args.hist_pooling,
        block_size=args.block_size,
        stride=args.block_size // args.stride_divider,
        valid_pixels_min_ratio=1 - args.max_saturated,
        normalization=args.histograms_normalization,
        use_sqrt=args.sqrt)
    max_pca_components = min(
        features_weighted.shape[0] * features_weighted.shape[1],
        features_weighted.shape[2])
    pca_n_components = min(
        pca_n_components,
        max_pca_components) if pca_n_components > 0 else max_pca_components
    if mode == 'mixture':
        mask_training = mask_valid
        mahalanobis_valid, mahalanobis_weighted, mahalanobis_full = predictor(
            features_weighted,
            features_full,
            mask_training,
            pca_n_components=pca_n_components,
            n_init=args.n_init,
            max_iter=args.max_iter,
            tol=args.tolerance,
            **kwargs_mixture)
    else:
        mask = mask[cuts_y.astype(int)][:, cuts_x.astype(int)]
        mask_training = mask_valid * ~mask
        mahalanobis_valid, mahalanobis_weighted, mahalanobis_full = predict_direct(
            features_weighted,
            features_full,
            mask_training,
            pca_n_components=pca_n_components)

    mahalanobis_valid = np.square(mahalanobis_valid)
    mahalanobis_weighted = np.square(mahalanobis_weighted) * mask_loose
    mahalanobis_full = np.square(mahalanobis_full)

    mahalanobis_valid = _to_fullsize(cuts_y,
                                     cuts_x,
                                     mahalanobis_valid,
                                     Y,
                                     X,
                                     fill=0)
    mahalanobis_weighted = _to_fullsize(cuts_y,
                                        cuts_x,
                                        mahalanobis_weighted,
                                        Y,
                                        X,
                                        fill=0)
    mahalanobis_full = _to_fullsize(cuts_y,
                                    cuts_x,
                                    mahalanobis_full,
                                    Y,
                                    X,
                                    fill=0)

    # full_mahal = mahalanobis.copy()
    # mahalanobis[
    # ~mask_valid] = 0  # remove saturated areas from the detection

    # mahalanobis = _to_fullsize(cuts_y, cuts_x, mahalanobis, Y, X, fill=0)
    # full_mahal = _to_fullsize(cuts_y, cuts_x, full_mahal, Y, X, fill=0)

    mask_valid = _to_fullsize(cuts_y,
                              cuts_x,
                              mask_valid,
                              Y,
                              X,
                              method='nearest')
    mask_training = _to_fullsize(cuts_y,
                                 cuts_x,
                                 mask_training,
                                 Y,
                                 X,
                                 method='nearest')

    # nfa = np.count_nonzero(mask_valid) * (
    # 1 - sp.stats.chi2.cdf(mahalanobis, df=pca_n_components))
    # nfa = nfa.clip(1e-12, 100)
    # binary = nfa < args.nfa

    out = args.out

    fig, ax = plt.subplots(figsize=FIGSIZE)
    draw = ax.matshow(mahalanobis_valid,
                      cmap='mako',
                      vmax=np.quantile(mahalanobis_valid, .98))
    cb = create_colorbar(draw, ax)
    ax.axis('off')
    plt.savefig(f'{out}_mahal_valid.png', bbox_inches='tight', pad_inches=.1)

    fig, ax = plt.subplots(figsize=FIGSIZE)
    draw = ax.matshow(mahalanobis_weighted,
                      cmap='mako',
                      vmax=np.quantile(mahalanobis_weighted, .98))
    cb = create_colorbar(draw, ax)
    ax.axis('off')
    plt.savefig(f'{out}_mahal_weighted.png',
                bbox_inches='tight',
                pad_inches=.1)

    fig, ax = plt.subplots(figsize=FIGSIZE)
    draw = ax.matshow(mahalanobis_full,
                      cmap='mako',
                      vmax=np.quantile(mahalanobis_full, .98))
    cb = create_colorbar(draw, ax)
    ax.axis('off')
    plt.savefig(f'{out}_mahal_full.png', bbox_inches='tight', pad_inches=.1)

    # save normalized distance
    min_, max_ = mahalanobis_valid.min(), mahalanobis_valid.max()
    mahalanobis_norm = 255 * (mahalanobis_valid - min_) / (max_ - min_)
    mahalanobis_norm = mahalanobis_norm.astype(np.uint8)
    imageio.imsave(f'{out}_normalized_distance.png', mahalanobis_norm)

    # save saturation mask
    imageio.imsave(f'{out}_mask_valid.png', mask_valid.astype(np.uint8) * 255)
    imageio.imsave(f'{out}_mask_loose.png', mask_loose.astype(np.uint8) * 255)

    # save training mask
    imageio.imsave(f'{out}_mask_training.png',
                   mask_training.astype(np.uint8) * 255)
